#ifndef _FUNCTIONTIMER_H_
#define _FUNCTIONTIMER_H_

#include "Exception.h"

#include <string>
#include <map>
#include <sys/time.h>

///////////////////////////////////////////////////////////////////////////////

class FunctionTimer {

public:
	///	<summary>
	///		Microseconds per second.
	///	</summary>
	static const unsigned long MICROSECONDS_PER_SECOND = 1000000;

public:
	///	<summary>
	///		A structure for storing group data.
	///	</summary>
	struct TimerGroupData {
		unsigned long iTotalTime;
		unsigned int nEntries;
	};

	///	<summary>
	///		The map structure for storing group data.
	///	</summary>
	typedef std::map<std::string, TimerGroupData> GroupDataMap;

	typedef std::pair<std::string, TimerGroupData> GroupDataPair;

public:
	///	<summary>
	///		Constructor.
	///	</summary>
	FunctionTimer(const char *szGroup = NULL) {

		// Assign group name
		if (szGroup == NULL) {
			m_strGroup = "";
		} else {
			m_strGroup = szGroup;
		}

		// Assign start time
		gettimeofday(&m_tvStartTime, NULL);
	}

	///	<summary>
	///		Reset the timer.
	///	</summary>
	void Reset() {
		gettimeofday(&m_tvStartTime, NULL);
	}

	///	<summary>
	///		Return the time elapsed since this timer began.
	///	</summary>
	///	<param name="fDone">
	///		If true stores the elapsed time in the group structure.
	///	</param>
	unsigned long Time(bool fDone = false) {
		timeval tv;
		gettimeofday(&tv, NULL);

		unsigned long iTime = 
		    MICROSECONDS_PER_SECOND * (tv.tv_sec - m_tvStartTime.tv_sec)
		    + (tv.tv_usec - m_tvStartTime.tv_usec);

		// If no name associated with this timer, ignore fDone.
		if (m_strGroup == "") {
			return iTime;
		}

		// Add the time to the group record
		if (fDone) {
			GroupDataMap::iterator iter;

			iter = m_mapGroupData.find(m_strGroup);

			// Add to existing group record
			if (iter != m_mapGroupData.end()) {
				iter->second.iTotalTime += iTime;
				iter->second.nEntries++;

			// Create new group record
			} else {
				GroupDataPair gdp;
				gdp.first = m_strGroup;
				gdp.second.iTotalTime = iTime;
				gdp.second.nEntries = 1;

				m_mapGroupData.insert(gdp);
			}
		}

		return iTime;
	}

	///	<summary>
	///		Return the time elapsed since this timer began and store in
	///		group data.
	///	</summary>
	unsigned long StopTime() {
		return Time(true);
	}

public:
	///	<summary>
	///		Receive a group data record.
	///	</summary>
	static const TimerGroupData & GetGroupTimeRecord(const char *szName) {
		GroupDataMap::iterator iter;

		iter = m_mapGroupData.find(szName);

		// Retrieve existing group record
		if (iter != m_mapGroupData.end()) {
			return iter->second;

		// Group record does not exist
		} else {
			_EXCEPTION1("Group time record %s does not exist.", szName);
		}
	}

	///	<summary>
	///		Retrieve the average time from a group data record.
	///	</summary>
	static unsigned long GetAverageGroupTime(const char *szName) {
		const TimerGroupData &tgd = GetGroupTimeRecord(szName);

		return (tgd.iTotalTime / tgd.nEntries);
	}

	///	<summary>
	///		Reset the group data record.
	///	</summary>
	static void ResetGroupTimeRecord(const char *szName) {
		GroupDataMap::iterator iter;

		iter = m_mapGroupData.find(szName);

		// Retrieve existing group record
		if (iter != m_mapGroupData.end()) {
			m_mapGroupData.erase(iter);

		// Group record does not exist
		} else {
			_EXCEPTION1("Group time record %s does not exist.", szName);
		}
	}

private:
	///	<summary>
	///		Group data.
	///	</summary>
	static GroupDataMap m_mapGroupData;

private:
	///	<summary>
	///		Time at which this timer was constructed.
	///	</summary
	timeval m_tvStartTime;

	///	<summary>
	///		Group name associated with this timer.
	///	</summary>
	std::string m_strGroup;
};

///////////////////////////////////////////////////////////////////////////////

#endif

